/* This file is generated by glib-genmarshal, do not modify it. This code is licensed under the same license as the containing project. Note that it links to GLib, so must comply with the LGPL linking clauses. */
#ifndef ___WOCKY_SIGNALS_MARSHAL_MARSHAL_H__
#define ___WOCKY_SIGNALS_MARSHAL_MARSHAL_H__

#include <glib-object.h>

G_BEGIN_DECLS

/* VOID:OBJECT (wocky-signals-marshal.list:1) */
#define _wocky_signals_marshal_VOID__OBJECT	g_cclosure_marshal_VOID__OBJECT

/* VOID:OBJECT,ENUM,STRING (wocky-signals-marshal.list:2) */
extern
void _wocky_signals_marshal_VOID__OBJECT_ENUM_STRING (GClosure     *closure,
                                                      GValue       *return_value,
                                                      guint         n_param_values,
                                                      const GValue *param_values,
                                                      gpointer      invocation_hint,
                                                      gpointer      marshal_data);

/* VOID:OBJECT,ENUM,STRING,LONG,POINTER,STRING,ENUM,ENUM (wocky-signals-marshal.list:3) */
extern
void _wocky_signals_marshal_VOID__OBJECT_ENUM_STRING_LONG_POINTER_STRING_ENUM_ENUM (GClosure     *closure,
                                                                                    GValue       *return_value,
                                                                                    guint         n_param_values,
                                                                                    const GValue *param_values,
                                                                                    gpointer      invocation_hint,
                                                                                    gpointer      marshal_data);

/* VOID:OBJECT,ENUM,STRING,LONG,POINTER,STRING,STRING,ENUM (wocky-signals-marshal.list:4) */
extern
void _wocky_signals_marshal_VOID__OBJECT_ENUM_STRING_LONG_POINTER_STRING_STRING_ENUM (GClosure     *closure,
                                                                                      GValue       *return_value,
                                                                                      guint         n_param_values,
                                                                                      const GValue *param_values,
                                                                                      gpointer      invocation_hint,
                                                                                      gpointer      marshal_data);

/* VOID:OBJECT,OBJECT,POINTER (wocky-signals-marshal.list:5) */
extern
void _wocky_signals_marshal_VOID__OBJECT_OBJECT_POINTER (GClosure     *closure,
                                                         GValue       *return_value,
                                                         guint         n_param_values,
                                                         const GValue *param_values,
                                                         gpointer      invocation_hint,
                                                         gpointer      marshal_data);

/* VOID:OBJECT,OBJECT,POINTER,POINTER (wocky-signals-marshal.list:6) */
extern
void _wocky_signals_marshal_VOID__OBJECT_OBJECT_POINTER_POINTER (GClosure     *closure,
                                                                 GValue       *return_value,
                                                                 guint         n_param_values,
                                                                 const GValue *param_values,
                                                                 gpointer      invocation_hint,
                                                                 gpointer      marshal_data);

/* VOID:OBJECT,OBJECT,POINTER,POINTER,BOXED (wocky-signals-marshal.list:7) */
extern
void _wocky_signals_marshal_VOID__OBJECT_OBJECT_POINTER_POINTER_BOXED (GClosure     *closure,
                                                                       GValue       *return_value,
                                                                       guint         n_param_values,
                                                                       const GValue *param_values,
                                                                       gpointer      invocation_hint,
                                                                       gpointer      marshal_data);

/* VOID:OBJECT,OBJECT,POINTER,POINTER,POINTER (wocky-signals-marshal.list:8) */
extern
void _wocky_signals_marshal_VOID__OBJECT_OBJECT_POINTER_POINTER_POINTER (GClosure     *closure,
                                                                         GValue       *return_value,
                                                                         guint         n_param_values,
                                                                         const GValue *param_values,
                                                                         gpointer      invocation_hint,
                                                                         gpointer      marshal_data);

/* VOID:OBJECT,POINTER,POINTER (wocky-signals-marshal.list:9) */
extern
void _wocky_signals_marshal_VOID__OBJECT_POINTER_POINTER (GClosure     *closure,
                                                          GValue       *return_value,
                                                          guint         n_param_values,
                                                          const GValue *param_values,
                                                          gpointer      invocation_hint,
                                                          gpointer      marshal_data);

/* VOID:OBJECT,POINTER,POINTER,BOXED (wocky-signals-marshal.list:10) */
extern
void _wocky_signals_marshal_VOID__OBJECT_POINTER_POINTER_BOXED (GClosure     *closure,
                                                                GValue       *return_value,
                                                                guint         n_param_values,
                                                                const GValue *param_values,
                                                                gpointer      invocation_hint,
                                                                gpointer      marshal_data);

/* VOID:OBJECT,POINTER,POINTER,POINTER (wocky-signals-marshal.list:11) */
extern
void _wocky_signals_marshal_VOID__OBJECT_POINTER_POINTER_POINTER (GClosure     *closure,
                                                                  GValue       *return_value,
                                                                  guint         n_param_values,
                                                                  const GValue *param_values,
                                                                  gpointer      invocation_hint,
                                                                  gpointer      marshal_data);

/* VOID:OBJECT,UINT (wocky-signals-marshal.list:12) */
extern
void _wocky_signals_marshal_VOID__OBJECT_UINT (GClosure     *closure,
                                               GValue       *return_value,
                                               guint         n_param_values,
                                               const GValue *param_values,
                                               gpointer      invocation_hint,
                                               gpointer      marshal_data);

/* VOID:OBJECT,UINT,POINTER (wocky-signals-marshal.list:13) */
extern
void _wocky_signals_marshal_VOID__OBJECT_UINT_POINTER (GClosure     *closure,
                                                       GValue       *return_value,
                                                       guint         n_param_values,
                                                       const GValue *param_values,
                                                       gpointer      invocation_hint,
                                                       gpointer      marshal_data);

/* VOID:OBJECT,UINT,POINTER,STRING,STRING,STRING (wocky-signals-marshal.list:14) */
extern
void _wocky_signals_marshal_VOID__OBJECT_UINT_POINTER_STRING_STRING_STRING (GClosure     *closure,
                                                                            GValue       *return_value,
                                                                            guint         n_param_values,
                                                                            const GValue *param_values,
                                                                            gpointer      invocation_hint,
                                                                            gpointer      marshal_data);

/* VOID:OBJECT,UINT,STRING,STRING,STRING (wocky-signals-marshal.list:15) */
extern
void _wocky_signals_marshal_VOID__OBJECT_UINT_STRING_STRING_STRING (GClosure     *closure,
                                                                    GValue       *return_value,
                                                                    guint         n_param_values,
                                                                    const GValue *param_values,
                                                                    gpointer      invocation_hint,
                                                                    gpointer      marshal_data);

/* VOID:POINTER,UINT (wocky-signals-marshal.list:16) */
extern
void _wocky_signals_marshal_VOID__POINTER_UINT (GClosure     *closure,
                                                GValue       *return_value,
                                                guint         n_param_values,
                                                const GValue *param_values,
                                                gpointer      invocation_hint,
                                                gpointer      marshal_data);

/* VOID:POINTER,UINT,POINTER,POINTER (wocky-signals-marshal.list:17) */
extern
void _wocky_signals_marshal_VOID__POINTER_UINT_POINTER_POINTER (GClosure     *closure,
                                                                GValue       *return_value,
                                                                guint         n_param_values,
                                                                const GValue *param_values,
                                                                gpointer      invocation_hint,
                                                                gpointer      marshal_data);

/* VOID:UINT,INT,STRING (wocky-signals-marshal.list:18) */
extern
void _wocky_signals_marshal_VOID__UINT_INT_STRING (GClosure     *closure,
                                                   GValue       *return_value,
                                                   guint         n_param_values,
                                                   const GValue *param_values,
                                                   gpointer      invocation_hint,
                                                   gpointer      marshal_data);


G_END_DECLS

#endif /* ___WOCKY_SIGNALS_MARSHAL_MARSHAL_H__ */
