
/* This file is generated by glib-mkenums, do not modify it. This code is licensed under the same license as the containing project. Note that it links to GLib, so must comply with the LGPL linking clauses. */

#include <wocky-xmpp-reader.h>
#include <wocky-xmpp-reader-enumtypes.h>

/* enumerations from "wocky-xmpp-reader.h" */
GType
wocky_xmpp_reader_state_get_type (void)
{
  static GType etype = 0;
  if (etype == 0) {
    static const GEnumValue values[] = {
      { WOCKY_XMPP_READER_STATE_INITIAL, "WOCKY_XMPP_READER_STATE_INITIAL", "initial" },
      { WOCKY_XMPP_READER_STATE_OPENED, "WOCKY_XMPP_READER_STATE_OPENED", "opened" },
      { WOCKY_XMPP_READER_STATE_CLOSED, "WOCKY_XMPP_READER_STATE_CLOSED", "closed" },
      { WOCKY_XMPP_READER_STATE_ERROR, "WOCKY_XMPP_READER_STATE_ERROR", "error" },
      { 0, NULL, NULL }
    };
    etype = g_enum_register_static ("WockyXmppReaderState", values);
  }
  return etype;
}
GType
wocky_xmpp_reader_error_get_type (void)
{
  static GType etype = 0;
  if (etype == 0) {
    static const GEnumValue values[] = {
      { WOCKY_XMPP_READER_ERROR_INVALID_STREAM_START, "WOCKY_XMPP_READER_ERROR_INVALID_STREAM_START", "invalid-stream-start" },
      { WOCKY_XMPP_READER_ERROR_PARSE_ERROR, "WOCKY_XMPP_READER_ERROR_PARSE_ERROR", "parse-error" },
      { 0, NULL, NULL }
    };
    etype = g_enum_register_static ("WockyXmppReaderError", values);
  }
  return etype;
}

/* Generated data ends here */

