
/* This file is generated by glib-mkenums, do not modify it. This code is licensed under the same license as the containing project. Note that it links to GLib, so must comply with the LGPL linking clauses. */

#include <wocky-data-form.h>
#include <wocky-data-form-enumtypes.h>

/* enumerations from "wocky-data-form.h" */
GType
wocky_data_form_field_type_get_type (void)
{
  static GType etype = 0;
  if (etype == 0) {
    static const GEnumValue values[] = {
      { WOCKY_DATA_FORM_FIELD_TYPE_BOOLEAN, "WOCKY_DATA_FORM_FIELD_TYPE_BOOLEAN", "boolean" },
      { WOCKY_DATA_FORM_FIELD_TYPE_FIXED, "WOCKY_DATA_FORM_FIELD_TYPE_FIXED", "fixed" },
      { WOCKY_DATA_FORM_FIELD_TYPE_HIDDEN, "WOCKY_DATA_FORM_FIELD_TYPE_HIDDEN", "hidden" },
      { WOCKY_DATA_FORM_FIELD_TYPE_JID_MULTI, "WOCKY_DATA_FORM_FIELD_TYPE_JID_MULTI", "jid-multi" },
      { WOCKY_DATA_FORM_FIELD_TYPE_JID_SINGLE, "WOCKY_DATA_FORM_FIELD_TYPE_JID_SINGLE", "jid-single" },
      { WOCKY_DATA_FORM_FIELD_TYPE_LIST_MULTI, "WOCKY_DATA_FORM_FIELD_TYPE_LIST_MULTI", "list-multi" },
      { WOCKY_DATA_FORM_FIELD_TYPE_LIST_SINGLE, "WOCKY_DATA_FORM_FIELD_TYPE_LIST_SINGLE", "list-single" },
      { WOCKY_DATA_FORM_FIELD_TYPE_TEXT_MULTI, "WOCKY_DATA_FORM_FIELD_TYPE_TEXT_MULTI", "text-multi" },
      { WOCKY_DATA_FORM_FIELD_TYPE_TEXT_PRIVATE, "WOCKY_DATA_FORM_FIELD_TYPE_TEXT_PRIVATE", "text-private" },
      { WOCKY_DATA_FORM_FIELD_TYPE_TEXT_SINGLE, "WOCKY_DATA_FORM_FIELD_TYPE_TEXT_SINGLE", "text-single" },
      { 0, NULL, NULL }
    };
    etype = g_enum_register_static ("WockyDataFormFieldType", values);
  }
  return etype;
}
GType
wocky_data_form_error_get_type (void)
{
  static GType etype = 0;
  if (etype == 0) {
    static const GEnumValue values[] = {
      { WOCKY_DATA_FORM_ERROR_NOT_FORM, "WOCKY_DATA_FORM_ERROR_NOT_FORM", "not-form" },
      { WOCKY_DATA_FORM_ERROR_WRONG_TYPE, "WOCKY_DATA_FORM_ERROR_WRONG_TYPE", "wrong-type" },
      { 0, NULL, NULL }
    };
    etype = g_enum_register_static ("WockyDataFormError", values);
  }
  return etype;
}

/* Generated data ends here */

