
/* This file is generated by glib-mkenums, do not modify it. This code is licensed under the same license as the containing project. Note that it links to GLib, so must comply with the LGPL linking clauses. */

#include <wocky-xmpp-error.h>
#include <wocky-xmpp-error-enumtypes.h>

/* enumerations from "wocky-xmpp-error.h" */
GType
wocky_xmpp_error_type_get_type (void)
{
  static GType etype = 0;
  if (etype == 0) {
    static const GEnumValue values[] = {
      { WOCKY_XMPP_ERROR_TYPE_CANCEL, "WOCKY_XMPP_ERROR_TYPE_CANCEL", "cancel" },
      { WOCKY_XMPP_ERROR_TYPE_CONTINUE, "WOCKY_XMPP_ERROR_TYPE_CONTINUE", "continue" },
      { WOCKY_XMPP_ERROR_TYPE_MODIFY, "WOCKY_XMPP_ERROR_TYPE_MODIFY", "modify" },
      { WOCKY_XMPP_ERROR_TYPE_AUTH, "WOCKY_XMPP_ERROR_TYPE_AUTH", "auth" },
      { WOCKY_XMPP_ERROR_TYPE_WAIT, "WOCKY_XMPP_ERROR_TYPE_WAIT", "wait" },
      { 0, NULL, NULL }
    };
    etype = g_enum_register_static ("WockyXmppErrorType", values);
  }
  return etype;
}
GType
wocky_xmpp_error_get_type (void)
{
  static GType etype = 0;
  if (etype == 0) {
    static const GEnumValue values[] = {
      { WOCKY_XMPP_ERROR_UNDEFINED_CONDITION, "WOCKY_XMPP_ERROR_UNDEFINED_CONDITION", "undefined-condition" },
      { WOCKY_XMPP_ERROR_REDIRECT, "WOCKY_XMPP_ERROR_REDIRECT", "redirect" },
      { WOCKY_XMPP_ERROR_GONE, "WOCKY_XMPP_ERROR_GONE", "gone" },
      { WOCKY_XMPP_ERROR_BAD_REQUEST, "WOCKY_XMPP_ERROR_BAD_REQUEST", "bad-request" },
      { WOCKY_XMPP_ERROR_UNEXPECTED_REQUEST, "WOCKY_XMPP_ERROR_UNEXPECTED_REQUEST", "unexpected-request" },
      { WOCKY_XMPP_ERROR_JID_MALFORMED, "WOCKY_XMPP_ERROR_JID_MALFORMED", "jid-malformed" },
      { WOCKY_XMPP_ERROR_NOT_AUTHORIZED, "WOCKY_XMPP_ERROR_NOT_AUTHORIZED", "not-authorized" },
      { WOCKY_XMPP_ERROR_PAYMENT_REQUIRED, "WOCKY_XMPP_ERROR_PAYMENT_REQUIRED", "payment-required" },
      { WOCKY_XMPP_ERROR_FORBIDDEN, "WOCKY_XMPP_ERROR_FORBIDDEN", "forbidden" },
      { WOCKY_XMPP_ERROR_ITEM_NOT_FOUND, "WOCKY_XMPP_ERROR_ITEM_NOT_FOUND", "item-not-found" },
      { WOCKY_XMPP_ERROR_RECIPIENT_UNAVAILABLE, "WOCKY_XMPP_ERROR_RECIPIENT_UNAVAILABLE", "recipient-unavailable" },
      { WOCKY_XMPP_ERROR_REMOTE_SERVER_NOT_FOUND, "WOCKY_XMPP_ERROR_REMOTE_SERVER_NOT_FOUND", "remote-server-not-found" },
      { WOCKY_XMPP_ERROR_NOT_ALLOWED, "WOCKY_XMPP_ERROR_NOT_ALLOWED", "not-allowed" },
      { WOCKY_XMPP_ERROR_NOT_ACCEPTABLE, "WOCKY_XMPP_ERROR_NOT_ACCEPTABLE", "not-acceptable" },
      { WOCKY_XMPP_ERROR_REGISTRATION_REQUIRED, "WOCKY_XMPP_ERROR_REGISTRATION_REQUIRED", "registration-required" },
      { WOCKY_XMPP_ERROR_SUBSCRIPTION_REQUIRED, "WOCKY_XMPP_ERROR_SUBSCRIPTION_REQUIRED", "subscription-required" },
      { WOCKY_XMPP_ERROR_REMOTE_SERVER_TIMEOUT, "WOCKY_XMPP_ERROR_REMOTE_SERVER_TIMEOUT", "remote-server-timeout" },
      { WOCKY_XMPP_ERROR_CONFLICT, "WOCKY_XMPP_ERROR_CONFLICT", "conflict" },
      { WOCKY_XMPP_ERROR_INTERNAL_SERVER_ERROR, "WOCKY_XMPP_ERROR_INTERNAL_SERVER_ERROR", "internal-server-error" },
      { WOCKY_XMPP_ERROR_RESOURCE_CONSTRAINT, "WOCKY_XMPP_ERROR_RESOURCE_CONSTRAINT", "resource-constraint" },
      { WOCKY_XMPP_ERROR_FEATURE_NOT_IMPLEMENTED, "WOCKY_XMPP_ERROR_FEATURE_NOT_IMPLEMENTED", "feature-not-implemented" },
      { WOCKY_XMPP_ERROR_SERVICE_UNAVAILABLE, "WOCKY_XMPP_ERROR_SERVICE_UNAVAILABLE", "service-unavailable" },
      { 0, NULL, NULL }
    };
    etype = g_enum_register_static ("WockyXmppError", values);
  }
  return etype;
}
GType
wocky_jingle_error_get_type (void)
{
  static GType etype = 0;
  if (etype == 0) {
    static const GEnumValue values[] = {
      { WOCKY_JINGLE_ERROR_OUT_OF_ORDER, "WOCKY_JINGLE_ERROR_OUT_OF_ORDER", "out-of-order" },
      { WOCKY_JINGLE_ERROR_TIE_BREAK, "WOCKY_JINGLE_ERROR_TIE_BREAK", "tie-break" },
      { WOCKY_JINGLE_ERROR_UNKNOWN_SESSION, "WOCKY_JINGLE_ERROR_UNKNOWN_SESSION", "unknown-session" },
      { WOCKY_JINGLE_ERROR_UNSUPPORTED_INFO, "WOCKY_JINGLE_ERROR_UNSUPPORTED_INFO", "unsupported-info" },
      { 0, NULL, NULL }
    };
    etype = g_enum_register_static ("WockyJingleError", values);
  }
  return etype;
}
GType
wocky_si_error_get_type (void)
{
  static GType etype = 0;
  if (etype == 0) {
    static const GEnumValue values[] = {
      { WOCKY_SI_ERROR_NO_VALID_STREAMS, "WOCKY_SI_ERROR_NO_VALID_STREAMS", "no-valid-streams" },
      { WOCKY_SI_ERROR_BAD_PROFILE, "WOCKY_SI_ERROR_BAD_PROFILE", "bad-profile" },
      { 0, NULL, NULL }
    };
    etype = g_enum_register_static ("WockySIError", values);
  }
  return etype;
}
GType
wocky_xmpp_stream_error_get_type (void)
{
  static GType etype = 0;
  if (etype == 0) {
    static const GEnumValue values[] = {
      { WOCKY_XMPP_STREAM_ERROR_BAD_FORMAT, "WOCKY_XMPP_STREAM_ERROR_BAD_FORMAT", "bad-format" },
      { WOCKY_XMPP_STREAM_ERROR_BAD_NAMESPACE_PREFIX, "WOCKY_XMPP_STREAM_ERROR_BAD_NAMESPACE_PREFIX", "bad-namespace-prefix" },
      { WOCKY_XMPP_STREAM_ERROR_CONFLICT, "WOCKY_XMPP_STREAM_ERROR_CONFLICT", "conflict" },
      { WOCKY_XMPP_STREAM_ERROR_CONNECTION_TIMEOUT, "WOCKY_XMPP_STREAM_ERROR_CONNECTION_TIMEOUT", "connection-timeout" },
      { WOCKY_XMPP_STREAM_ERROR_HOST_GONE, "WOCKY_XMPP_STREAM_ERROR_HOST_GONE", "host-gone" },
      { WOCKY_XMPP_STREAM_ERROR_HOST_UNKNOWN, "WOCKY_XMPP_STREAM_ERROR_HOST_UNKNOWN", "host-unknown" },
      { WOCKY_XMPP_STREAM_ERROR_IMPROPER_ADDRESSING, "WOCKY_XMPP_STREAM_ERROR_IMPROPER_ADDRESSING", "improper-addressing" },
      { WOCKY_XMPP_STREAM_ERROR_INTERNAL_SERVER_ERROR, "WOCKY_XMPP_STREAM_ERROR_INTERNAL_SERVER_ERROR", "internal-server-error" },
      { WOCKY_XMPP_STREAM_ERROR_INVALID_FROM, "WOCKY_XMPP_STREAM_ERROR_INVALID_FROM", "invalid-from" },
      { WOCKY_XMPP_STREAM_ERROR_INVALID_ID, "WOCKY_XMPP_STREAM_ERROR_INVALID_ID", "invalid-id" },
      { WOCKY_XMPP_STREAM_ERROR_INVALID_NAMESPACE, "WOCKY_XMPP_STREAM_ERROR_INVALID_NAMESPACE", "invalid-namespace" },
      { WOCKY_XMPP_STREAM_ERROR_INVALID_XML, "WOCKY_XMPP_STREAM_ERROR_INVALID_XML", "invalid-xml" },
      { WOCKY_XMPP_STREAM_ERROR_NOT_AUTHORIZED, "WOCKY_XMPP_STREAM_ERROR_NOT_AUTHORIZED", "not-authorized" },
      { WOCKY_XMPP_STREAM_ERROR_POLICY_VIOLATION, "WOCKY_XMPP_STREAM_ERROR_POLICY_VIOLATION", "policy-violation" },
      { WOCKY_XMPP_STREAM_ERROR_REMOTE_CONNECTION_FAILED, "WOCKY_XMPP_STREAM_ERROR_REMOTE_CONNECTION_FAILED", "remote-connection-failed" },
      { WOCKY_XMPP_STREAM_ERROR_RESOURCE_CONSTRAINT, "WOCKY_XMPP_STREAM_ERROR_RESOURCE_CONSTRAINT", "resource-constraint" },
      { WOCKY_XMPP_STREAM_ERROR_RESTRICTED_XML, "WOCKY_XMPP_STREAM_ERROR_RESTRICTED_XML", "restricted-xml" },
      { WOCKY_XMPP_STREAM_ERROR_SEE_OTHER_HOST, "WOCKY_XMPP_STREAM_ERROR_SEE_OTHER_HOST", "see-other-host" },
      { WOCKY_XMPP_STREAM_ERROR_SYSTEM_SHUTDOWN, "WOCKY_XMPP_STREAM_ERROR_SYSTEM_SHUTDOWN", "system-shutdown" },
      { WOCKY_XMPP_STREAM_ERROR_UNDEFINED_CONDITION, "WOCKY_XMPP_STREAM_ERROR_UNDEFINED_CONDITION", "undefined-condition" },
      { WOCKY_XMPP_STREAM_ERROR_UNSUPPORTED_ENCODING, "WOCKY_XMPP_STREAM_ERROR_UNSUPPORTED_ENCODING", "unsupported-encoding" },
      { WOCKY_XMPP_STREAM_ERROR_UNSUPPORTED_STANZA_TYPE, "WOCKY_XMPP_STREAM_ERROR_UNSUPPORTED_STANZA_TYPE", "unsupported-stanza-type" },
      { WOCKY_XMPP_STREAM_ERROR_UNSUPPORTED_VERSION, "WOCKY_XMPP_STREAM_ERROR_UNSUPPORTED_VERSION", "unsupported-version" },
      { WOCKY_XMPP_STREAM_ERROR_XML_NOT_WELL_FORMED, "WOCKY_XMPP_STREAM_ERROR_XML_NOT_WELL_FORMED", "xml-not-well-formed" },
      { WOCKY_XMPP_STREAM_ERROR_UNKNOWN, "WOCKY_XMPP_STREAM_ERROR_UNKNOWN", "unknown" },
      { 0, NULL, NULL }
    };
    etype = g_enum_register_static ("WockyXmppStreamError", values);
  }
  return etype;
}

/* Generated data ends here */

