#include <glib-object.h>
#include <dbus/dbus-glib.h>
#include <telepathy-glib/dbus-properties-mixin.h>

G_BEGIN_DECLS

/**
 * SalutSvcConnectionFUTURE:
 *
 * Dummy typedef representing any implementation of this interface.
 */
typedef struct _SalutSvcConnectionFUTURE SalutSvcConnectionFUTURE;

/**
 * SalutSvcConnectionFUTUREClass:
 *
 * The class of SalutSvcConnectionFUTURE.
 */
typedef struct _SalutSvcConnectionFUTUREClass SalutSvcConnectionFUTUREClass;

GType salut_svc_connection_future_get_type (void);
#define SALUT_TYPE_SVC_CONNECTION_FUTURE \
  (salut_svc_connection_future_get_type ())
#define SALUT_SVC_CONNECTION_FUTURE(obj) \
  (G_TYPE_CHECK_INSTANCE_CAST((obj), SALUT_TYPE_SVC_CONNECTION_FUTURE, SalutSvcConnectionFUTURE))
#define SALUT_IS_SVC_CONNECTION_FUTURE(obj) \
  (G_TYPE_CHECK_INSTANCE_TYPE((obj), SALUT_TYPE_SVC_CONNECTION_FUTURE))
#define SALUT_SVC_CONNECTION_FUTURE_GET_CLASS(obj) \
  (G_TYPE_INSTANCE_GET_INTERFACE((obj), SALUT_TYPE_SVC_CONNECTION_FUTURE, SalutSvcConnectionFUTUREClass))


typedef void (*salut_svc_connection_future_ensure_sidecar_impl) (SalutSvcConnectionFUTURE *self,
    const gchar *in_Main_Interface,
    DBusGMethodInvocation *context);
void salut_svc_connection_future_implement_ensure_sidecar (SalutSvcConnectionFUTUREClass *klass, salut_svc_connection_future_ensure_sidecar_impl impl);
/**
 * salut_svc_connection_future_return_from_ensure_sidecar:
 * @context: The D-Bus method invocation context
 * @out_Path: const gchar * (FIXME, generate documentation)
 * @out_Properties: GHashTable * (FIXME, generate documentation)
 *
 * Return successfully by calling dbus_g_method_return().
 * This inline function exists only to provide type-safety.
 */
static inline
/* this comment is to stop gtkdoc realising this is static */
void salut_svc_connection_future_return_from_ensure_sidecar (DBusGMethodInvocation *context,
    const gchar *out_Path,
    GHashTable *out_Properties);
static inline void
salut_svc_connection_future_return_from_ensure_sidecar (DBusGMethodInvocation *context,
    const gchar *out_Path,
    GHashTable *out_Properties)
{
  dbus_g_method_return (context,
      out_Path,
      out_Properties);
}


/**
 * SalutSvcOLPCActivityProperties:
 *
 * Dummy typedef representing any implementation of this interface.
 */
typedef struct _SalutSvcOLPCActivityProperties SalutSvcOLPCActivityProperties;

/**
 * SalutSvcOLPCActivityPropertiesClass:
 *
 * The class of SalutSvcOLPCActivityProperties.
 */
typedef struct _SalutSvcOLPCActivityPropertiesClass SalutSvcOLPCActivityPropertiesClass;

GType salut_svc_olpc_activity_properties_get_type (void);
#define SALUT_TYPE_SVC_OLPC_ACTIVITY_PROPERTIES \
  (salut_svc_olpc_activity_properties_get_type ())
#define SALUT_SVC_OLPC_ACTIVITY_PROPERTIES(obj) \
  (G_TYPE_CHECK_INSTANCE_CAST((obj), SALUT_TYPE_SVC_OLPC_ACTIVITY_PROPERTIES, SalutSvcOLPCActivityProperties))
#define SALUT_IS_SVC_OLPC_ACTIVITY_PROPERTIES(obj) \
  (G_TYPE_CHECK_INSTANCE_TYPE((obj), SALUT_TYPE_SVC_OLPC_ACTIVITY_PROPERTIES))
#define SALUT_SVC_OLPC_ACTIVITY_PROPERTIES_GET_CLASS(obj) \
  (G_TYPE_INSTANCE_GET_INTERFACE((obj), SALUT_TYPE_SVC_OLPC_ACTIVITY_PROPERTIES, SalutSvcOLPCActivityPropertiesClass))


typedef void (*salut_svc_olpc_activity_properties_set_properties_impl) (SalutSvcOLPCActivityProperties *self,
    guint in_room,
    GHashTable *in_properties,
    DBusGMethodInvocation *context);
void salut_svc_olpc_activity_properties_implement_set_properties (SalutSvcOLPCActivityPropertiesClass *klass, salut_svc_olpc_activity_properties_set_properties_impl impl);
/**
 * salut_svc_olpc_activity_properties_return_from_set_properties:
 * @context: The D-Bus method invocation context
 *
 * Return successfully by calling dbus_g_method_return().
 * This inline function exists only to provide type-safety.
 */
static inline
/* this comment is to stop gtkdoc realising this is static */
void salut_svc_olpc_activity_properties_return_from_set_properties (DBusGMethodInvocation *context);
static inline void
salut_svc_olpc_activity_properties_return_from_set_properties (DBusGMethodInvocation *context)
{
  dbus_g_method_return (context);
}

typedef void (*salut_svc_olpc_activity_properties_get_properties_impl) (SalutSvcOLPCActivityProperties *self,
    guint in_room,
    DBusGMethodInvocation *context);
void salut_svc_olpc_activity_properties_implement_get_properties (SalutSvcOLPCActivityPropertiesClass *klass, salut_svc_olpc_activity_properties_get_properties_impl impl);
/**
 * salut_svc_olpc_activity_properties_return_from_get_properties:
 * @context: The D-Bus method invocation context
 * @out_properties: GHashTable * (FIXME, generate documentation)
 *
 * Return successfully by calling dbus_g_method_return().
 * This inline function exists only to provide type-safety.
 */
static inline
/* this comment is to stop gtkdoc realising this is static */
void salut_svc_olpc_activity_properties_return_from_get_properties (DBusGMethodInvocation *context,
    GHashTable *out_properties);
static inline void
salut_svc_olpc_activity_properties_return_from_get_properties (DBusGMethodInvocation *context,
    GHashTable *out_properties)
{
  dbus_g_method_return (context,
      out_properties);
}

typedef void (*salut_svc_olpc_activity_properties_get_activity_impl) (SalutSvcOLPCActivityProperties *self,
    const gchar *in_activity_id,
    DBusGMethodInvocation *context);
void salut_svc_olpc_activity_properties_implement_get_activity (SalutSvcOLPCActivityPropertiesClass *klass, salut_svc_olpc_activity_properties_get_activity_impl impl);
/**
 * salut_svc_olpc_activity_properties_return_from_get_activity:
 * @context: The D-Bus method invocation context
 * @out_room: guint  (FIXME, generate documentation)
 *
 * Return successfully by calling dbus_g_method_return().
 * This inline function exists only to provide type-safety.
 */
static inline
/* this comment is to stop gtkdoc realising this is static */
void salut_svc_olpc_activity_properties_return_from_get_activity (DBusGMethodInvocation *context,
    guint out_room);
static inline void
salut_svc_olpc_activity_properties_return_from_get_activity (DBusGMethodInvocation *context,
    guint out_room)
{
  dbus_g_method_return (context,
      out_room);
}

void salut_svc_olpc_activity_properties_emit_activity_properties_changed (gpointer instance,
    guint arg_room,
    GHashTable *arg_properties);

/**
 * SalutSvcOLPCBuddyInfo:
 *
 * Dummy typedef representing any implementation of this interface.
 */
typedef struct _SalutSvcOLPCBuddyInfo SalutSvcOLPCBuddyInfo;

/**
 * SalutSvcOLPCBuddyInfoClass:
 *
 * The class of SalutSvcOLPCBuddyInfo.
 */
typedef struct _SalutSvcOLPCBuddyInfoClass SalutSvcOLPCBuddyInfoClass;

GType salut_svc_olpc_buddy_info_get_type (void);
#define SALUT_TYPE_SVC_OLPC_BUDDY_INFO \
  (salut_svc_olpc_buddy_info_get_type ())
#define SALUT_SVC_OLPC_BUDDY_INFO(obj) \
  (G_TYPE_CHECK_INSTANCE_CAST((obj), SALUT_TYPE_SVC_OLPC_BUDDY_INFO, SalutSvcOLPCBuddyInfo))
#define SALUT_IS_SVC_OLPC_BUDDY_INFO(obj) \
  (G_TYPE_CHECK_INSTANCE_TYPE((obj), SALUT_TYPE_SVC_OLPC_BUDDY_INFO))
#define SALUT_SVC_OLPC_BUDDY_INFO_GET_CLASS(obj) \
  (G_TYPE_INSTANCE_GET_INTERFACE((obj), SALUT_TYPE_SVC_OLPC_BUDDY_INFO, SalutSvcOLPCBuddyInfoClass))


typedef void (*salut_svc_olpc_buddy_info_set_properties_impl) (SalutSvcOLPCBuddyInfo *self,
    GHashTable *in_properties,
    DBusGMethodInvocation *context);
void salut_svc_olpc_buddy_info_implement_set_properties (SalutSvcOLPCBuddyInfoClass *klass, salut_svc_olpc_buddy_info_set_properties_impl impl);
/**
 * salut_svc_olpc_buddy_info_return_from_set_properties:
 * @context: The D-Bus method invocation context
 *
 * Return successfully by calling dbus_g_method_return().
 * This inline function exists only to provide type-safety.
 */
static inline
/* this comment is to stop gtkdoc realising this is static */
void salut_svc_olpc_buddy_info_return_from_set_properties (DBusGMethodInvocation *context);
static inline void
salut_svc_olpc_buddy_info_return_from_set_properties (DBusGMethodInvocation *context)
{
  dbus_g_method_return (context);
}

typedef void (*salut_svc_olpc_buddy_info_get_properties_impl) (SalutSvcOLPCBuddyInfo *self,
    guint in_contact,
    DBusGMethodInvocation *context);
void salut_svc_olpc_buddy_info_implement_get_properties (SalutSvcOLPCBuddyInfoClass *klass, salut_svc_olpc_buddy_info_get_properties_impl impl);
/**
 * salut_svc_olpc_buddy_info_return_from_get_properties:
 * @context: The D-Bus method invocation context
 * @out_properties: GHashTable * (FIXME, generate documentation)
 *
 * Return successfully by calling dbus_g_method_return().
 * This inline function exists only to provide type-safety.
 */
static inline
/* this comment is to stop gtkdoc realising this is static */
void salut_svc_olpc_buddy_info_return_from_get_properties (DBusGMethodInvocation *context,
    GHashTable *out_properties);
static inline void
salut_svc_olpc_buddy_info_return_from_get_properties (DBusGMethodInvocation *context,
    GHashTable *out_properties)
{
  dbus_g_method_return (context,
      out_properties);
}

typedef void (*salut_svc_olpc_buddy_info_set_activities_impl) (SalutSvcOLPCBuddyInfo *self,
    const GPtrArray *in_activities,
    DBusGMethodInvocation *context);
void salut_svc_olpc_buddy_info_implement_set_activities (SalutSvcOLPCBuddyInfoClass *klass, salut_svc_olpc_buddy_info_set_activities_impl impl);
/**
 * salut_svc_olpc_buddy_info_return_from_set_activities:
 * @context: The D-Bus method invocation context
 *
 * Return successfully by calling dbus_g_method_return().
 * This inline function exists only to provide type-safety.
 */
static inline
/* this comment is to stop gtkdoc realising this is static */
void salut_svc_olpc_buddy_info_return_from_set_activities (DBusGMethodInvocation *context);
static inline void
salut_svc_olpc_buddy_info_return_from_set_activities (DBusGMethodInvocation *context)
{
  dbus_g_method_return (context);
}

typedef void (*salut_svc_olpc_buddy_info_add_activity_impl) (SalutSvcOLPCBuddyInfo *self,
    const gchar *in_id,
    guint in_handle,
    DBusGMethodInvocation *context);
void salut_svc_olpc_buddy_info_implement_add_activity (SalutSvcOLPCBuddyInfoClass *klass, salut_svc_olpc_buddy_info_add_activity_impl impl);
/**
 * salut_svc_olpc_buddy_info_return_from_add_activity:
 * @context: The D-Bus method invocation context
 *
 * Return successfully by calling dbus_g_method_return().
 * This inline function exists only to provide type-safety.
 */
static inline
/* this comment is to stop gtkdoc realising this is static */
void salut_svc_olpc_buddy_info_return_from_add_activity (DBusGMethodInvocation *context);
static inline void
salut_svc_olpc_buddy_info_return_from_add_activity (DBusGMethodInvocation *context)
{
  dbus_g_method_return (context);
}

typedef void (*salut_svc_olpc_buddy_info_get_activities_impl) (SalutSvcOLPCBuddyInfo *self,
    guint in_contact,
    DBusGMethodInvocation *context);
void salut_svc_olpc_buddy_info_implement_get_activities (SalutSvcOLPCBuddyInfoClass *klass, salut_svc_olpc_buddy_info_get_activities_impl impl);
/**
 * salut_svc_olpc_buddy_info_return_from_get_activities:
 * @context: The D-Bus method invocation context
 * @out_activities: const GPtrArray * (FIXME, generate documentation)
 *
 * Return successfully by calling dbus_g_method_return().
 * This inline function exists only to provide type-safety.
 */
static inline
/* this comment is to stop gtkdoc realising this is static */
void salut_svc_olpc_buddy_info_return_from_get_activities (DBusGMethodInvocation *context,
    const GPtrArray *out_activities);
static inline void
salut_svc_olpc_buddy_info_return_from_get_activities (DBusGMethodInvocation *context,
    const GPtrArray *out_activities)
{
  dbus_g_method_return (context,
      out_activities);
}

typedef void (*salut_svc_olpc_buddy_info_set_current_activity_impl) (SalutSvcOLPCBuddyInfo *self,
    const gchar *in_activity,
    guint in_channel,
    DBusGMethodInvocation *context);
void salut_svc_olpc_buddy_info_implement_set_current_activity (SalutSvcOLPCBuddyInfoClass *klass, salut_svc_olpc_buddy_info_set_current_activity_impl impl);
/**
 * salut_svc_olpc_buddy_info_return_from_set_current_activity:
 * @context: The D-Bus method invocation context
 *
 * Return successfully by calling dbus_g_method_return().
 * This inline function exists only to provide type-safety.
 */
static inline
/* this comment is to stop gtkdoc realising this is static */
void salut_svc_olpc_buddy_info_return_from_set_current_activity (DBusGMethodInvocation *context);
static inline void
salut_svc_olpc_buddy_info_return_from_set_current_activity (DBusGMethodInvocation *context)
{
  dbus_g_method_return (context);
}

typedef void (*salut_svc_olpc_buddy_info_get_current_activity_impl) (SalutSvcOLPCBuddyInfo *self,
    guint in_contact,
    DBusGMethodInvocation *context);
void salut_svc_olpc_buddy_info_implement_get_current_activity (SalutSvcOLPCBuddyInfoClass *klass, salut_svc_olpc_buddy_info_get_current_activity_impl impl);
/**
 * salut_svc_olpc_buddy_info_return_from_get_current_activity:
 * @context: The D-Bus method invocation context
 * @out_activity: const gchar * (FIXME, generate documentation)
 * @out_channel: guint  (FIXME, generate documentation)
 *
 * Return successfully by calling dbus_g_method_return().
 * This inline function exists only to provide type-safety.
 */
static inline
/* this comment is to stop gtkdoc realising this is static */
void salut_svc_olpc_buddy_info_return_from_get_current_activity (DBusGMethodInvocation *context,
    const gchar *out_activity,
    guint out_channel);
static inline void
salut_svc_olpc_buddy_info_return_from_get_current_activity (DBusGMethodInvocation *context,
    const gchar *out_activity,
    guint out_channel)
{
  dbus_g_method_return (context,
      out_activity,
      out_channel);
}

void salut_svc_olpc_buddy_info_emit_properties_changed (gpointer instance,
    guint arg_contact,
    GHashTable *arg_properties);
void salut_svc_olpc_buddy_info_emit_activities_changed (gpointer instance,
    guint arg_contact,
    const GPtrArray *arg_activities);
void salut_svc_olpc_buddy_info_emit_current_activity_changed (gpointer instance,
    guint arg_contact,
    const gchar *arg_activity,
    guint arg_channel);

/**
 * SalutSvcSalutPluginTest:
 *
 * Dummy typedef representing any implementation of this interface.
 */
typedef struct _SalutSvcSalutPluginTest SalutSvcSalutPluginTest;

/**
 * SalutSvcSalutPluginTestClass:
 *
 * The class of SalutSvcSalutPluginTest.
 */
typedef struct _SalutSvcSalutPluginTestClass SalutSvcSalutPluginTestClass;

GType salut_svc_salut_plugin_test_get_type (void);
#define SALUT_TYPE_SVC_SALUT_PLUGIN_TEST \
  (salut_svc_salut_plugin_test_get_type ())
#define SALUT_SVC_SALUT_PLUGIN_TEST(obj) \
  (G_TYPE_CHECK_INSTANCE_CAST((obj), SALUT_TYPE_SVC_SALUT_PLUGIN_TEST, SalutSvcSalutPluginTest))
#define SALUT_IS_SVC_SALUT_PLUGIN_TEST(obj) \
  (G_TYPE_CHECK_INSTANCE_TYPE((obj), SALUT_TYPE_SVC_SALUT_PLUGIN_TEST))
#define SALUT_SVC_SALUT_PLUGIN_TEST_GET_CLASS(obj) \
  (G_TYPE_INSTANCE_GET_INTERFACE((obj), SALUT_TYPE_SVC_SALUT_PLUGIN_TEST, SalutSvcSalutPluginTestClass))




G_END_DECLS
