
/* This file is generated by glib-mkenums, do not modify it. This code is licensed under the same license as the containing project. Note that it links to GLib, so must comply with the LGPL linking clauses. */

#include <wocky-muc.h>
#include <wocky-muc-enumtypes.h>

/* enumerations from "wocky-muc.h" */
GType
wocky_muc_status_code_get_type (void)
{
  static GType etype = 0;
  if (etype == 0) {
    static const GFlagsValue values[] = {
      { WOCKY_MUC_CODE_UNKNOWN, "WOCKY_MUC_CODE_UNKNOWN", "unknown" },
      { WOCKY_MUC_CODE_ONYMOUS, "WOCKY_MUC_CODE_ONYMOUS", "onymous" },
      { WOCKY_MUC_CODE_AF_CHANGE_OOB, "WOCKY_MUC_CODE_AF_CHANGE_OOB", "af-change-oob" },
      { WOCKY_MUC_CODE_CFG_SHOW_UNAVAILABLE, "WOCKY_MUC_CODE_CFG_SHOW_UNAVAILABLE", "cfg-show-unavailable" },
      { WOCKY_MUC_CODE_CFG_HIDE_UNAVAILABLE, "WOCKY_MUC_CODE_CFG_HIDE_UNAVAILABLE", "cfg-hide-unavailable" },
      { WOCKY_MUC_CODE_CFG_NONPRIVACY, "WOCKY_MUC_CODE_CFG_NONPRIVACY", "cfg-nonprivacy" },
      { WOCKY_MUC_CODE_OWN_PRESENCE, "WOCKY_MUC_CODE_OWN_PRESENCE", "own-presence" },
      { WOCKY_MUC_CODE_CFG_LOGGING_ENABLED, "WOCKY_MUC_CODE_CFG_LOGGING_ENABLED", "cfg-logging-enabled" },
      { WOCKY_MUC_CODE_CFG_LOGGING_DISABLED, "WOCKY_MUC_CODE_CFG_LOGGING_DISABLED", "cfg-logging-disabled" },
      { WOCKY_MUC_CODE_CFG_ONYMOUS, "WOCKY_MUC_CODE_CFG_ONYMOUS", "cfg-onymous" },
      { WOCKY_MUC_CODE_CFG_SEMIONYMOUS, "WOCKY_MUC_CODE_CFG_SEMIONYMOUS", "cfg-semionymous" },
      { WOCKY_MUC_CODE_CFG_ANONYMOUS, "WOCKY_MUC_CODE_CFG_ANONYMOUS", "cfg-anonymous" },
      { WOCKY_MUC_CODE_NEW_ROOM, "WOCKY_MUC_CODE_NEW_ROOM", "new-room" },
      { WOCKY_MUC_CODE_NICK_CHANGE_FORCED, "WOCKY_MUC_CODE_NICK_CHANGE_FORCED", "nick-change-forced" },
      { WOCKY_MUC_CODE_BANNED, "WOCKY_MUC_CODE_BANNED", "banned" },
      { WOCKY_MUC_CODE_NICK_CHANGE_USER, "WOCKY_MUC_CODE_NICK_CHANGE_USER", "nick-change-user" },
      { WOCKY_MUC_CODE_KICKED, "WOCKY_MUC_CODE_KICKED", "kicked" },
      { WOCKY_MUC_CODE_KICKED_AFFILIATION, "WOCKY_MUC_CODE_KICKED_AFFILIATION", "kicked-affiliation" },
      { WOCKY_MUC_CODE_KICKED_ROOM_PRIVATISED, "WOCKY_MUC_CODE_KICKED_ROOM_PRIVATISED", "kicked-room-privatised" },
      { WOCKY_MUC_CODE_KICKED_SHUTDOWN, "WOCKY_MUC_CODE_KICKED_SHUTDOWN", "kicked-shutdown" },
      { 0, NULL, NULL }
    };
    etype = g_flags_register_static ("WockyMucStatusCode", values);
  }
  return etype;
}
GType
wocky_muc_role_get_type (void)
{
  static GType etype = 0;
  if (etype == 0) {
    static const GEnumValue values[] = {
      { WOCKY_MUC_ROLE_NONE, "WOCKY_MUC_ROLE_NONE", "none" },
      { WOCKY_MUC_ROLE_VISITOR, "WOCKY_MUC_ROLE_VISITOR", "visitor" },
      { WOCKY_MUC_ROLE_PARTICIPANT, "WOCKY_MUC_ROLE_PARTICIPANT", "participant" },
      { WOCKY_MUC_ROLE_MODERATOR, "WOCKY_MUC_ROLE_MODERATOR", "moderator" },
      { 0, NULL, NULL }
    };
    etype = g_enum_register_static ("WockyMucRole", values);
  }
  return etype;
}
GType
wocky_muc_affiliation_get_type (void)
{
  static GType etype = 0;
  if (etype == 0) {
    static const GEnumValue values[] = {
      { WOCKY_MUC_AFFILIATION_OUTCAST, "WOCKY_MUC_AFFILIATION_OUTCAST", "outcast" },
      { WOCKY_MUC_AFFILIATION_NONE, "WOCKY_MUC_AFFILIATION_NONE", "none" },
      { WOCKY_MUC_AFFILIATION_MEMBER, "WOCKY_MUC_AFFILIATION_MEMBER", "member" },
      { WOCKY_MUC_AFFILIATION_ADMIN, "WOCKY_MUC_AFFILIATION_ADMIN", "admin" },
      { WOCKY_MUC_AFFILIATION_OWNER, "WOCKY_MUC_AFFILIATION_OWNER", "owner" },
      { 0, NULL, NULL }
    };
    etype = g_enum_register_static ("WockyMucAffiliation", values);
  }
  return etype;
}
GType
wocky_muc_feature_get_type (void)
{
  static GType etype = 0;
  if (etype == 0) {
    static const GFlagsValue values[] = {
      { WOCKY_MUC_MODERN, "WOCKY_MUC_MODERN", "modern" },
      { WOCKY_MUC_FORM_REGISTER, "WOCKY_MUC_FORM_REGISTER", "form-register" },
      { WOCKY_MUC_FORM_ROOMCONFIG, "WOCKY_MUC_FORM_ROOMCONFIG", "form-roomconfig" },
      { WOCKY_MUC_FORM_ROOMINFO, "WOCKY_MUC_FORM_ROOMINFO", "form-roominfo" },
      { WOCKY_MUC_HIDDEN, "WOCKY_MUC_HIDDEN", "hidden" },
      { WOCKY_MUC_MEMBERSONLY, "WOCKY_MUC_MEMBERSONLY", "membersonly" },
      { WOCKY_MUC_MODERATED, "WOCKY_MUC_MODERATED", "moderated" },
      { WOCKY_MUC_NONANONYMOUS, "WOCKY_MUC_NONANONYMOUS", "nonanonymous" },
      { WOCKY_MUC_OPEN, "WOCKY_MUC_OPEN", "open" },
      { WOCKY_MUC_PASSWORDPROTECTED, "WOCKY_MUC_PASSWORDPROTECTED", "passwordprotected" },
      { WOCKY_MUC_PERSISTENT, "WOCKY_MUC_PERSISTENT", "persistent" },
      { WOCKY_MUC_PUBLIC, "WOCKY_MUC_PUBLIC", "public" },
      { WOCKY_MUC_ROOMS, "WOCKY_MUC_ROOMS", "rooms" },
      { WOCKY_MUC_SEMIANONYMOUS, "WOCKY_MUC_SEMIANONYMOUS", "semianonymous" },
      { WOCKY_MUC_TEMPORARY, "WOCKY_MUC_TEMPORARY", "temporary" },
      { WOCKY_MUC_UNMODERATED, "WOCKY_MUC_UNMODERATED", "unmoderated" },
      { WOCKY_MUC_UNSECURED, "WOCKY_MUC_UNSECURED", "unsecured" },
      { WOCKY_MUC_OBSOLETE, "WOCKY_MUC_OBSOLETE", "obsolete" },
      { 0, NULL, NULL }
    };
    etype = g_flags_register_static ("WockyMucFeature", values);
  }
  return etype;
}
GType
wocky_muc_msg_type_get_type (void)
{
  static GType etype = 0;
  if (etype == 0) {
    static const GEnumValue values[] = {
      { WOCKY_MUC_MSG_NONE, "WOCKY_MUC_MSG_NONE", "none" },
      { WOCKY_MUC_MSG_NORMAL, "WOCKY_MUC_MSG_NORMAL", "normal" },
      { WOCKY_MUC_MSG_ACTION, "WOCKY_MUC_MSG_ACTION", "action" },
      { WOCKY_MUC_MSG_NOTICE, "WOCKY_MUC_MSG_NOTICE", "notice" },
      { 0, NULL, NULL }
    };
    etype = g_enum_register_static ("WockyMucMsgType", values);
  }
  return etype;
}
GType
wocky_muc_msg_state_get_type (void)
{
  static GType etype = 0;
  if (etype == 0) {
    static const GEnumValue values[] = {
      { WOCKY_MUC_MSG_STATE_NONE, "WOCKY_MUC_MSG_STATE_NONE", "none" },
      { WOCKY_MUC_MSG_STATE_ACTIVE, "WOCKY_MUC_MSG_STATE_ACTIVE", "active" },
      { WOCKY_MUC_MSG_STATE_COMPOSING, "WOCKY_MUC_MSG_STATE_COMPOSING", "composing" },
      { WOCKY_MUC_MSG_STATE_INACTIVE, "WOCKY_MUC_MSG_STATE_INACTIVE", "inactive" },
      { WOCKY_MUC_MSG_STATE_PAUSED, "WOCKY_MUC_MSG_STATE_PAUSED", "paused" },
      { 0, NULL, NULL }
    };
    etype = g_enum_register_static ("WockyMucMsgState", values);
  }
  return etype;
}
GType
wocky_muc_state_get_type (void)
{
  static GType etype = 0;
  if (etype == 0) {
    static const GEnumValue values[] = {
      { WOCKY_MUC_CREATED, "WOCKY_MUC_CREATED", "created" },
      { WOCKY_MUC_INITIATED, "WOCKY_MUC_INITIATED", "initiated" },
      { WOCKY_MUC_AUTH, "WOCKY_MUC_AUTH", "auth" },
      { WOCKY_MUC_JOINED, "WOCKY_MUC_JOINED", "joined" },
      { WOCKY_MUC_ENDED, "WOCKY_MUC_ENDED", "ended" },
      { 0, NULL, NULL }
    };
    etype = g_enum_register_static ("WockyMucState", values);
  }
  return etype;
}

/* Generated data ends here */

