
/* This file is generated by glib-mkenums, do not modify it. This code is licensed under the same license as the containing project. Note that it links to GLib, so must comply with the LGPL linking clauses. */

#ifndef __WOCKY_XMPP_ERROR_ENUM_TYPES_H__
#define __WOCKY_XMPP_ERROR_ENUM_TYPES_H__

#include <glib-object.h>

G_BEGIN_DECLS
/* enumerations from "wocky-xmpp-error.h" */
GType wocky_xmpp_error_type_get_type (void);
#define WOCKY_TYPE_XMPP_ERROR_TYPE (wocky_xmpp_error_type_get_type())
GType wocky_xmpp_error_get_type (void);
#define WOCKY_TYPE_XMPP_ERROR (wocky_xmpp_error_get_type())
GType wocky_jingle_error_get_type (void);
#define WOCKY_TYPE_JINGLE_ERROR (wocky_jingle_error_get_type())
GType wocky_si_error_get_type (void);
#define WOCKY_TYPE_SI_ERROR (wocky_si_error_get_type())
GType wocky_xmpp_stream_error_get_type (void);
#define WOCKY_TYPE_XMPP_STREAM_ERROR (wocky_xmpp_stream_error_get_type())
G_END_DECLS

#endif

/* Generated data ends here */

