
/* This file is generated by glib-mkenums, do not modify it. This code is licensed under the same license as the containing project. Note that it links to GLib, so must comply with the LGPL linking clauses. */

#ifndef __WOCKY_PUBSUB_NODE_ENUM_TYPES_H__
#define __WOCKY_PUBSUB_NODE_ENUM_TYPES_H__

#include <glib-object.h>

G_BEGIN_DECLS
/* enumerations from "wocky-pubsub-node.h" */
GType wocky_pubsub_affiliation_state_get_type (void);
#define WOCKY_TYPE_PUBSUB_AFFILIATION_STATE (wocky_pubsub_affiliation_state_get_type())
G_END_DECLS

#endif

/* Generated data ends here */

