/* This file is generated by glib-genmarshal, do not modify it. This code is licensed under the same license as the containing project. Note that it links to GLib, so must comply with the LGPL linking clauses. */
#ifndef __SALUT_SIGNALS_MARSHAL_MARSHAL_H__
#define __SALUT_SIGNALS_MARSHAL_MARSHAL_H__

#include <glib-object.h>

G_BEGIN_DECLS

/* VOID:OBJECT (signals-marshal.list:1) */
#define salut_signals_marshal_VOID__OBJECT	g_cclosure_marshal_VOID__OBJECT

/* VOID:OBJECT,INT (signals-marshal.list:2) */
extern
void salut_signals_marshal_VOID__OBJECT_INT (GClosure     *closure,
                                             GValue       *return_value,
                                             guint         n_param_values,
                                             const GValue *param_values,
                                             gpointer      invocation_hint,
                                             gpointer      marshal_data);

/* VOID:UINT (signals-marshal.list:3) */
#define salut_signals_marshal_VOID__UINT	g_cclosure_marshal_VOID__UINT

/* VOID:VOID (signals-marshal.list:4) */
#define salut_signals_marshal_VOID__VOID	g_cclosure_marshal_VOID__VOID


G_END_DECLS

#endif /* __SALUT_SIGNALS_MARSHAL_MARSHAL_H__ */
