
/* This file is generated by glib-mkenums, do not modify it. This code is licensed under the same license as the containing project. Note that it links to GLib, so must comply with the LGPL linking clauses. */

#include <wocky-pubsub-service.h>
#include <wocky-pubsub-service-enumtypes.h>

/* enumerations from "wocky-pubsub-service.h" */
GType
wocky_pubsub_service_error_get_type (void)
{
  static GType etype = 0;
  if (etype == 0) {
    static const GEnumValue values[] = {
      { WOCKY_PUBSUB_SERVICE_ERROR_WRONG_REPLY, "WOCKY_PUBSUB_SERVICE_ERROR_WRONG_REPLY", "reply" },
      { 0, NULL, NULL }
    };
    etype = g_enum_register_static ("WockyPubsubServiceError", values);
  }
  return etype;
}
GType
wocky_pubsub_subscription_state_get_type (void)
{
  static GType etype = 0;
  if (etype == 0) {
    static const GEnumValue values[] = {
      { WOCKY_PUBSUB_SUBSCRIPTION_NONE, "WOCKY_PUBSUB_SUBSCRIPTION_NONE", "none" },
      { WOCKY_PUBSUB_SUBSCRIPTION_PENDING, "WOCKY_PUBSUB_SUBSCRIPTION_PENDING", "pending" },
      { WOCKY_PUBSUB_SUBSCRIPTION_SUBSCRIBED, "WOCKY_PUBSUB_SUBSCRIPTION_SUBSCRIBED", "subscribed" },
      { WOCKY_PUBSUB_SUBSCRIPTION_UNCONFIGURED, "WOCKY_PUBSUB_SUBSCRIPTION_UNCONFIGURED", "unconfigured" },
      { 0, NULL, NULL }
    };
    etype = g_enum_register_static ("WockyPubsubSubscriptionState", values);
  }
  return etype;
}

/* Generated data ends here */

