
/* This file is generated by glib-mkenums, do not modify it. This code is licensed under the same license as the containing project. Note that it links to GLib, so must comply with the LGPL linking clauses. */

#ifndef __WOCKY_MUC_ENUM_TYPES_H__
#define __WOCKY_MUC_ENUM_TYPES_H__

#include <glib-object.h>

G_BEGIN_DECLS
/* enumerations from "wocky-muc.h" */
GType wocky_muc_status_code_get_type (void);
#define WOCKY_TYPE_MUC_STATUS_CODE (wocky_muc_status_code_get_type())
GType wocky_muc_role_get_type (void);
#define WOCKY_TYPE_MUC_ROLE (wocky_muc_role_get_type())
GType wocky_muc_affiliation_get_type (void);
#define WOCKY_TYPE_MUC_AFFILIATION (wocky_muc_affiliation_get_type())
GType wocky_muc_feature_get_type (void);
#define WOCKY_TYPE_MUC_FEATURE (wocky_muc_feature_get_type())
GType wocky_muc_msg_type_get_type (void);
#define WOCKY_TYPE_MUC_MSG_TYPE (wocky_muc_msg_type_get_type())
GType wocky_muc_msg_state_get_type (void);
#define WOCKY_TYPE_MUC_MSG_STATE (wocky_muc_msg_state_get_type())
GType wocky_muc_state_get_type (void);
#define WOCKY_TYPE_MUC_STATE (wocky_muc_state_get_type())
G_END_DECLS

#endif

/* Generated data ends here */

