
/* This file is generated by glib-mkenums, do not modify it. This code is licensed under the same license as the containing project. Note that it links to GLib, so must comply with the LGPL linking clauses. */

#include <wocky-connector.h>
#include <wocky-connector-enumtypes.h>

/* enumerations from "wocky-connector.h" */
GType
wocky_connector_error_get_type (void)
{
  static GType etype = 0;
  if (etype == 0) {
    static const GEnumValue values[] = {
      { WOCKY_CONNECTOR_ERROR_UNKNOWN, "WOCKY_CONNECTOR_ERROR_UNKNOWN", "unknown" },
      { WOCKY_CONNECTOR_ERROR_IN_PROGRESS, "WOCKY_CONNECTOR_ERROR_IN_PROGRESS", "in-progress" },
      { WOCKY_CONNECTOR_ERROR_BAD_JID, "WOCKY_CONNECTOR_ERROR_BAD_JID", "bad-jid" },
      { WOCKY_CONNECTOR_ERROR_NON_XMPP_V1_SERVER, "WOCKY_CONNECTOR_ERROR_NON_XMPP_V1_SERVER", "non-xmpp-v1-server" },
      { WOCKY_CONNECTOR_ERROR_BAD_FEATURES, "WOCKY_CONNECTOR_ERROR_BAD_FEATURES", "bad-features" },
      { WOCKY_CONNECTOR_ERROR_TLS_UNAVAILABLE, "WOCKY_CONNECTOR_ERROR_TLS_UNAVAILABLE", "tls-unavailable" },
      { WOCKY_CONNECTOR_ERROR_TLS_REFUSED, "WOCKY_CONNECTOR_ERROR_TLS_REFUSED", "tls-refused" },
      { WOCKY_CONNECTOR_ERROR_TLS_SESSION_FAILED, "WOCKY_CONNECTOR_ERROR_TLS_SESSION_FAILED", "tls-session-failed" },
      { WOCKY_CONNECTOR_ERROR_BIND_UNAVAILABLE, "WOCKY_CONNECTOR_ERROR_BIND_UNAVAILABLE", "bind-unavailable" },
      { WOCKY_CONNECTOR_ERROR_BIND_FAILED, "WOCKY_CONNECTOR_ERROR_BIND_FAILED", "bind-failed" },
      { WOCKY_CONNECTOR_ERROR_BIND_INVALID, "WOCKY_CONNECTOR_ERROR_BIND_INVALID", "bind-invalid" },
      { WOCKY_CONNECTOR_ERROR_BIND_DENIED, "WOCKY_CONNECTOR_ERROR_BIND_DENIED", "bind-denied" },
      { WOCKY_CONNECTOR_ERROR_BIND_CONFLICT, "WOCKY_CONNECTOR_ERROR_BIND_CONFLICT", "bind-conflict" },
      { WOCKY_CONNECTOR_ERROR_BIND_REJECTED, "WOCKY_CONNECTOR_ERROR_BIND_REJECTED", "bind-rejected" },
      { WOCKY_CONNECTOR_ERROR_SESSION_FAILED, "WOCKY_CONNECTOR_ERROR_SESSION_FAILED", "session-failed" },
      { WOCKY_CONNECTOR_ERROR_SESSION_DENIED, "WOCKY_CONNECTOR_ERROR_SESSION_DENIED", "session-denied" },
      { WOCKY_CONNECTOR_ERROR_SESSION_CONFLICT, "WOCKY_CONNECTOR_ERROR_SESSION_CONFLICT", "session-conflict" },
      { WOCKY_CONNECTOR_ERROR_SESSION_REJECTED, "WOCKY_CONNECTOR_ERROR_SESSION_REJECTED", "session-rejected" },
      { WOCKY_CONNECTOR_ERROR_INSECURE, "WOCKY_CONNECTOR_ERROR_INSECURE", "insecure" },
      { WOCKY_CONNECTOR_ERROR_REGISTRATION_FAILED, "WOCKY_CONNECTOR_ERROR_REGISTRATION_FAILED", "registration-failed" },
      { WOCKY_CONNECTOR_ERROR_REGISTRATION_UNAVAILABLE, "WOCKY_CONNECTOR_ERROR_REGISTRATION_UNAVAILABLE", "registration-unavailable" },
      { WOCKY_CONNECTOR_ERROR_REGISTRATION_UNSUPPORTED, "WOCKY_CONNECTOR_ERROR_REGISTRATION_UNSUPPORTED", "registration-unsupported" },
      { WOCKY_CONNECTOR_ERROR_REGISTRATION_EMPTY, "WOCKY_CONNECTOR_ERROR_REGISTRATION_EMPTY", "registration-empty" },
      { WOCKY_CONNECTOR_ERROR_REGISTRATION_CONFLICT, "WOCKY_CONNECTOR_ERROR_REGISTRATION_CONFLICT", "registration-conflict" },
      { WOCKY_CONNECTOR_ERROR_REGISTRATION_REJECTED, "WOCKY_CONNECTOR_ERROR_REGISTRATION_REJECTED", "registration-rejected" },
      { WOCKY_CONNECTOR_ERROR_UNREGISTER_FAILED, "WOCKY_CONNECTOR_ERROR_UNREGISTER_FAILED", "unregister-failed" },
      { WOCKY_CONNECTOR_ERROR_UNREGISTER_DENIED, "WOCKY_CONNECTOR_ERROR_UNREGISTER_DENIED", "unregister-denied" },
      { 0, NULL, NULL }
    };
    etype = g_enum_register_static ("WockyConnectorError", values);
  }
  return etype;
}

/* Generated data ends here */

