/* This file is generated by glib-genmarshal, do not modify it. This code is licensed under the same license as the containing project. Note that it links to GLib, so must comply with the LGPL linking clauses. */
#ifndef ___SALUT_EXT_MARSHAL_MARSHAL_H__
#define ___SALUT_EXT_MARSHAL_MARSHAL_H__

#include <glib-object.h>

G_BEGIN_DECLS

/* VOID:BOXED,POINTER (_gen/signals-marshal.list:1) */
extern
void _salut_ext_marshal_VOID__BOXED_POINTER (GClosure     *closure,
                                             GValue       *return_value,
                                             guint         n_param_values,
                                             const GValue *param_values,
                                             gpointer      invocation_hint,
                                             gpointer      marshal_data);

/* VOID:STRING,UINT,POINTER (_gen/signals-marshal.list:2) */
extern
void _salut_ext_marshal_VOID__STRING_UINT_POINTER (GClosure     *closure,
                                                   GValue       *return_value,
                                                   guint         n_param_values,
                                                   const GValue *param_values,
                                                   gpointer      invocation_hint,
                                                   gpointer      marshal_data);

/* VOID:UINT,BOXED,POINTER (_gen/signals-marshal.list:3) */
extern
void _salut_ext_marshal_VOID__UINT_BOXED_POINTER (GClosure     *closure,
                                                  GValue       *return_value,
                                                  guint         n_param_values,
                                                  const GValue *param_values,
                                                  gpointer      invocation_hint,
                                                  gpointer      marshal_data);

/* VOID:STRING,POINTER (_gen/signals-marshal.list:4) */
extern
void _salut_ext_marshal_VOID__STRING_POINTER (GClosure     *closure,
                                              GValue       *return_value,
                                              guint         n_param_values,
                                              const GValue *param_values,
                                              gpointer      invocation_hint,
                                              gpointer      marshal_data);

/* VOID:UINT,BOXED (_gen/signals-marshal.list:5) */
extern
void _salut_ext_marshal_VOID__UINT_BOXED (GClosure     *closure,
                                          GValue       *return_value,
                                          guint         n_param_values,
                                          const GValue *param_values,
                                          gpointer      invocation_hint,
                                          gpointer      marshal_data);

/* VOID:UINT,STRING,UINT (_gen/signals-marshal.list:6) */
extern
void _salut_ext_marshal_VOID__UINT_STRING_UINT (GClosure     *closure,
                                                GValue       *return_value,
                                                guint         n_param_values,
                                                const GValue *param_values,
                                                gpointer      invocation_hint,
                                                gpointer      marshal_data);


G_END_DECLS

#endif /* ___SALUT_EXT_MARSHAL_MARSHAL_H__ */
