
/* This file is generated by glib-mkenums, do not modify it. This code is licensed under the same license as the containing project. Note that it links to GLib, so must comply with the LGPL linking clauses. */

#include <wocky-pubsub-node.h>
#include <wocky-pubsub-node-enumtypes.h>

/* enumerations from "wocky-pubsub-node.h" */
GType
wocky_pubsub_affiliation_state_get_type (void)
{
  static GType etype = 0;
  if (etype == 0) {
    static const GEnumValue values[] = {
      { WOCKY_PUBSUB_AFFILIATION_OWNER, "WOCKY_PUBSUB_AFFILIATION_OWNER", "owner" },
      { WOCKY_PUBSUB_AFFILIATION_PUBLISHER, "WOCKY_PUBSUB_AFFILIATION_PUBLISHER", "publisher" },
      { WOCKY_PUBSUB_AFFILIATION_PUBLISH_ONLY, "WOCKY_PUBSUB_AFFILIATION_PUBLISH_ONLY", "publish-only" },
      { WOCKY_PUBSUB_AFFILIATION_MEMBER, "WOCKY_PUBSUB_AFFILIATION_MEMBER", "member" },
      { WOCKY_PUBSUB_AFFILIATION_NONE, "WOCKY_PUBSUB_AFFILIATION_NONE", "none" },
      { WOCKY_PUBSUB_AFFILIATION_OUTCAST, "WOCKY_PUBSUB_AFFILIATION_OUTCAST", "outcast" },
      { 0, NULL, NULL }
    };
    etype = g_enum_register_static ("WockyPubsubAffiliationState", values);
  }
  return etype;
}

/* Generated data ends here */

