
/* This file is generated by glib-mkenums, do not modify it. This code is licensed under the same license as the containing project. Note that it links to GLib, so must comply with the LGPL linking clauses. */

#include <wocky-auth-registry.h>
#include <wocky-auth-registry-enumtypes.h>

/* enumerations from "wocky-auth-registry.h" */
GType
wocky_auth_error_get_type (void)
{
  static GType etype = 0;
  if (etype == 0) {
    static const GEnumValue values[] = {
      { WOCKY_AUTH_ERROR_INIT_FAILED, "WOCKY_AUTH_ERROR_INIT_FAILED", "init-failed" },
      { WOCKY_AUTH_ERROR_NOT_SUPPORTED, "WOCKY_AUTH_ERROR_NOT_SUPPORTED", "not-supported" },
      { WOCKY_AUTH_ERROR_NO_SUPPORTED_MECHANISMS, "WOCKY_AUTH_ERROR_NO_SUPPORTED_MECHANISMS", "no-supported-mechanisms" },
      { WOCKY_AUTH_ERROR_NETWORK, "WOCKY_AUTH_ERROR_NETWORK", "network" },
      { WOCKY_AUTH_ERROR_INVALID_REPLY, "WOCKY_AUTH_ERROR_INVALID_REPLY", "invalid-reply" },
      { WOCKY_AUTH_ERROR_NO_CREDENTIALS, "WOCKY_AUTH_ERROR_NO_CREDENTIALS", "no-credentials" },
      { WOCKY_AUTH_ERROR_FAILURE, "WOCKY_AUTH_ERROR_FAILURE", "failure" },
      { WOCKY_AUTH_ERROR_CONNRESET, "WOCKY_AUTH_ERROR_CONNRESET", "connreset" },
      { WOCKY_AUTH_ERROR_STREAM, "WOCKY_AUTH_ERROR_STREAM", "stream" },
      { WOCKY_AUTH_ERROR_RESOURCE_CONFLICT, "WOCKY_AUTH_ERROR_RESOURCE_CONFLICT", "resource-conflict" },
      { WOCKY_AUTH_ERROR_NOT_AUTHORIZED, "WOCKY_AUTH_ERROR_NOT_AUTHORIZED", "not-authorized" },
      { 0, NULL, NULL }
    };
    etype = g_enum_register_static ("WockyAuthError", values);
  }
  return etype;
}

/* Generated data ends here */

