/* This file is generated by glib-genmarshal, do not modify it. This code is licensed under the same license as the containing project. Note that it links to GLib, so must comply with the LGPL linking clauses. */
#ifndef ___GIBBER_SIGNALS_MARSHAL_MARSHAL_H__
#define ___GIBBER_SIGNALS_MARSHAL_MARSHAL_H__

#include <glib-object.h>

G_BEGIN_DECLS

/* VOID:OBJECT,OBJECT (gibber-signals-marshal.list:1) */
extern
void _gibber_signals_marshal_VOID__OBJECT_OBJECT (GClosure     *closure,
                                                  GValue       *return_value,
                                                  guint         n_param_values,
                                                  const GValue *param_values,
                                                  gpointer      invocation_hint,
                                                  gpointer      marshal_data);

/* VOID:OBJECT,POINTER,UINT (gibber-signals-marshal.list:2) */
extern
void _gibber_signals_marshal_VOID__OBJECT_POINTER_UINT (GClosure     *closure,
                                                        GValue       *return_value,
                                                        guint         n_param_values,
                                                        const GValue *param_values,
                                                        gpointer      invocation_hint,
                                                        gpointer      marshal_data);

/* VOID:STRING,OBJECT (gibber-signals-marshal.list:3) */
extern
void _gibber_signals_marshal_VOID__STRING_OBJECT (GClosure     *closure,
                                                  GValue       *return_value,
                                                  guint         n_param_values,
                                                  const GValue *param_values,
                                                  gpointer      invocation_hint,
                                                  gpointer      marshal_data);

/* VOID:STRING,POINTER (gibber-signals-marshal.list:4) */
extern
void _gibber_signals_marshal_VOID__STRING_POINTER (GClosure     *closure,
                                                   GValue       *return_value,
                                                   guint         n_param_values,
                                                   const GValue *param_values,
                                                   gpointer      invocation_hint,
                                                   gpointer      marshal_data);

/* VOID:STRING,UINT,POINTER,ULONG (gibber-signals-marshal.list:5) */
extern
void _gibber_signals_marshal_VOID__STRING_UINT_POINTER_ULONG (GClosure     *closure,
                                                              GValue       *return_value,
                                                              guint         n_param_values,
                                                              const GValue *param_values,
                                                              gpointer      invocation_hint,
                                                              gpointer      marshal_data);

/* VOID:UINT,INT,STRING (gibber-signals-marshal.list:6) */
extern
void _gibber_signals_marshal_VOID__UINT_INT_STRING (GClosure     *closure,
                                                    GValue       *return_value,
                                                    guint         n_param_values,
                                                    const GValue *param_values,
                                                    gpointer      invocation_hint,
                                                    gpointer      marshal_data);

/* VOID:UINT,POINTER,ULONG (gibber-signals-marshal.list:7) */
extern
void _gibber_signals_marshal_VOID__UINT_POINTER_ULONG (GClosure     *closure,
                                                       GValue       *return_value,
                                                       guint         n_param_values,
                                                       const GValue *param_values,
                                                       gpointer      invocation_hint,
                                                       gpointer      marshal_data);


G_END_DECLS

#endif /* ___GIBBER_SIGNALS_MARSHAL_MARSHAL_H__ */
